local addonName, addon = ...

-- functions
local ipairs, pairs, type, unpack, wipe = ipairs, pairs, type, unpack, wipe
local format = string.format
local tconcat, tinsert, tsort = table.concat, table.insert, table.sort
local ChatEdit_GetActiveWindow = ChatEdit_GetActiveWindow
local ChatEdit_InsertLink = ChatEdit_InsertLink
local ChatFrame_OpenChat = ChatFrame_OpenChat
local CreateFrame = CreateFrame
local GetAutoCompleteRealms = GetAutoCompleteRealms
local GetMessageTypeColor = GetMessageTypeColor
local IsInGroup = IsInGroup
local IsInGuild = IsInGuild
local IsShiftKeyDown = IsShiftKeyDown
-- local SendChatMessage = SendChatMessage
-- constants
local DEFAULT_CHAT_FRAME = DEFAULT_CHAT_FRAME
local LE_PARTY_CATEGORY_INSTANCE = LE_PARTY_CATEGORY_INSTANCE
-- frames
local UIParent = UIParent
-- libs
local LibStub = LibStub
local LibDataBroker = LibStub:GetLibrary("LibDataBroker-1.1")



-- workaround for issue #8
local SendChatMessage_G = SendChatMessage
local SendChatMessage_TextSplitLength = 240
local function SendChatMessage(...)
    local text = select(1, ...)

    if text and #text > SendChatMessage_TextSplitLength then
        local newArgs = {...}

        local leftPart = (text:sub(0, SendChatMessage_TextSplitLength) or ""):match("^%s*(.*%S)") or ""
        local rightPart = (text:sub(SendChatMessage_TextSplitLength + 1) or ""):match("^%s*(.*%S)") or ""

        if leftPart and #leftPart > 0 then
            newArgs[1] = leftPart
            SendChatMessage(unpack(newArgs))
        end

        if rightPart and #rightPart > 0 then
            newArgs[1] = rightPart
            SendChatMessage(unpack(newArgs))
        end

        return
    end

    SendChatMessage_G(...)
end


local broker
local private = {}



private.realms = {}
function private:GetCurrentRealms()
    wipe(private.realms)

    for _, realm in ipairs(GetAutoCompleteRealms()) do
        private.realms[realm] = true
    end

    private.realms[addon:GetRealmName()] = true

    return private.realms
end

function private:GetEligibleCharactersIterator(filter)
    local realms = self:GetCurrentRealms()

    return addon:GetCharactersIterator(function(_, c)
        return filter(_, c)
            and (addon.options.showOtherRealms == true or realms[c.realm])
            and (addon.options.showOtherFactions == true or c.faction == addon.player.faction)
    end)
end

private.keystones = {}
function private:BuildKeystoneTable(callback)
    wipe(self.keystones)
    for key, character in self:GetEligibleCharactersIterator(function(_, c) return c.keystone ~= nil end) do
        tinsert(self.keystones, format("%s (%s)",
            addon:GetKeystoneDisplayText(character.keystone),
            addon.options.anonymiseCharacters
                and addon:GetCharacterAnonymousDisplayName(key)
                or addon:GetCharacterDisplayName(key)
        ))
    end

    if #self.keystones > 0 then
        tsort(self.keystones)
        return callback(nil, self.keystones)
    end

    return callback("No keystones to report.")
end

function private:BuildKeystoneList(callback)
    return self:BuildKeystoneTable(function(err, tbl)
        if err then return callback(err) end

        return callback(nil, tconcat(tbl, ", "))
    end)
end

private.records = {}
function private:BuildRecordTable(callback)
    wipe(self.records)
    for key, character in self:GetEligibleCharactersIterator(function(_, c) return c.record ~= nil end) do
        local mapId, level = unpack(character.record)
        tinsert(self.records, format("%s %d (%s)",
            addon:GetMapName(mapId),
            level,
            addon.options.anonymiseCharacters
                and addon:GetCharacterAnonymousDisplayName(key)
                or addon:GetCharacterDisplayName(key)
        ))
    end

    if #private.records > 0 then
        tsort(self.records)
        return callback(nil, self.records)
    end

    return callback("No records to report.")
end

function private:BuildRecordList(callback)
    return self:BuildRecordTable(function(err, tbl)
        if err then return callback(err) end

        return callback(nil, tconcat(tbl, ", "))
    end)
end

function private:AddLinkToChat(link)
    local chatWindow = ChatEdit_GetActiveWindow()
    if chatWindow then
        ChatEdit_InsertLink(link)
    else
        ChatFrame_OpenChat(link, DEFAULT_CHAT_FRAME)
    end
end

function private:AddTextToChat(text)
    local chatWindow = ChatEdit_GetActiveWindow()
    if chatWindow then
        chatWindow:Insert(text)
    else
        ChatFrame_OpenChat(text, DEFAULT_CHAT_FRAME)
    end
end

function private:OnLeftClick()
    if IsShiftKeyDown() then
        if addon.player.keystone == nil then
            addon:Print("You don't have a keystone on this character.")
            addon:Print("If you did, shift-clicking the broker would link it to your chat.")

            return
        end

        return self:AddLinkToChat(addon.player.keystone)
    end

    local guildR, guildG, guildB = GetMessageTypeColor("GUILD")
    local guildHex = format("%x%x%x", guildR * 255, guildG * 255, guildB * 255)
    local partyR, partyG, partyB = GetMessageTypeColor("PARTY")
    local partyHex = format("%x%x%x", partyR * 255, partyG * 255, partyB * 255)

    local contextMenu = broker.contextMenu
    local item = contextMenu.item
    contextMenu:Show({
        item:Title("Report Keystones"),
        item:Action("in Chat Box", function()
            self:BuildKeystoneList(function(err, list)
                if err then addon:Print(err) return end

                self:AddTextToChat(format("Keystones: %s", list))
            end)
        end),
        item:Action(format("|cff%s%s|r", guildHex, "to Guild"), function()
            self:BuildKeystoneList(function(err, list)
                if err then addon:Print(err) return end

                if IsInGuild() then
                    SendChatMessage(format("Keystones: %s", list), "GUILD")
                else
                    addon:Print("You are not in a guild.")
                end
            end)
        end),
        item:Action(format("|cff%s%s|r", partyHex, "to Group"), function()
            self:BuildKeystoneList(function(err, list)
                if err then addon:Print(err) return end

                if IsInGroup(LE_PARTY_CATEGORY_INSTANCE) and not IsShiftKeyDown() then
                    SendChatMessage(format("Keystones: %s", list), "INSTANCE_CHAT")
                elseif IsInGroup() then
                    SendChatMessage(format("Keystones: %s", list), "PARTY")
                else
                    addon:Print("You are not in a party or instance.")
                end
            end)
        end),
        item:Separator(),
        item:Title("Report Best Runs"),
        item:Action("in Chat Box", function()
            self:BuildRecordList(function(err, list)
                if err then addon:Print(err) return end

                self:AddTextToChat(format("Best runs: %s", list))
            end)
        end),
        item:Action(format("|cff%s%s|r", guildHex, "to Guild"), function()
            self:BuildRecordList(function(err, list)
                if err then addon:Print(err) return end

                if IsInGuild() then
                    SendChatMessage(format("Best runs: %s", list), "GUILD")
                else
                    addon:Print("You are not in a guild.")
                end
            end)
        end),
        item:Action(format("|cff%s%s|r", partyHex, "to Group"), function()
            self:BuildRecordList(function(err, list)
                if err then addon:Print(err) return end

                if IsInGroup(LE_PARTY_CATEGORY_INSTANCE) and not IsShiftKeyDown() then
                    SendChatMessage(format("Best runs: %s", list), "INSTANCE_CHAT")
                elseif IsInGroup() then
                    SendChatMessage(format("Best runs: %s", list), "PARTY")
                else
                    addon:Print("You are not in a party or instance.")
                end
            end)
        end),
    })
end

function private:OnRightClick()
    local contextMenu = broker.contextMenu
    local item = contextMenu.item
    contextMenu:Show({
        item:Title(addonName),
        item:SubMenu("Options", {
            item:Title("General"),
            item:OptionToggle(
                "Use shorthand dungeon names where available",
                "useShorthand"
            ),
            item:OptionToggle(
                "Purge keystone and record data when weekly reset is detected (prompt otherwise)",
                "autoResetKeystones"
            ),
            item:OptionToggle(
                "Automatically slot keystones into font",
                "autoSlotKeystones"
            ),
            item:Separator(),
            item:Title("Broker"),
            item:OptionToggle(
                "Colour keystone levels by difficulty",
                "colourKeystoneLevelsByDifficulty"
            ),
            item:OptionToggle(
                "Show best run in broker text",
                "showRecordsInBrokerText"
            ),
            item:Checkbox(
                "Show minimap button",
                function() return addon:IsMinimapButtonShown() end,
                function() addon:ToggleMinimapButton() end
            ),
            item:Separator(),
            item:Title("Tooltip"),
            item:OptionToggle(
                "Show characters' best runs in tooltip",
                "showRecords"
            ),
            item:OptionToggle(
                "Show current affixes in tooltip",
                "showAffixes"
            ),
            item:OptionToggle(
                "Use classic-style tooltip",
                "useClassicTooltip"
            ),
            item:Separator(),
            item:Title("Data Reporting"),
            item:OptionToggle(
                "Use characters' class names instead of their real names",
                "anonymiseCharacters"
            ),
            item:OptionToggle(
                "Include characters from other realms",
                "showOtherRealms"
            ),
            item:OptionToggle(
                "Include characters from other factions",
                "showOtherFactions"
            ),
        }),
        item:SubMenu("Miscellaneous", {
            item:OptionToggle(
                "Debug mode",
                "debug"
            ),
            item:Separator(),
            item:Action("Delete Keystone Data (no confirmation)", function()
                addon:DeleteAllCharacterKeystones()
            end),
            item:Action("Delete Weekly Record Data (no confirmation)", function()
                addon:DeleteAllCharacterRecords()
            end),
        }),
        addon.options.debug and item:SubMenu("Debug", {
            item:Action("Trigger keystone reset prompt", function()
                addon:TriggerResetPrompt()
            end),
        }) or {},
    })
end



broker = LibDataBroker:NewDataObject(addonName, {
    type = "data source",
    label = "Keystone",
    text = "No keystone",
    icon = [[Interface\Icons\inv_relics_hourglass]],
})

addon:RegisterCallback(addon.EVENT_INITIALISED, function()
    addon:Fire(addon.EVENT_BROKER_INITIALISED, broker)
end)

function broker:OnClick(btn)
    if btn == "LeftButton" then return private:OnLeftClick() end
    if btn == "RightButton" then return private:OnRightClick() end
end

local function UpdateBrokerText()
    local character = addon:GetCharacter()
    if not character or not character.keystone then
        if character.previousBestRun then
            broker.text = format("|cffFF0000**|r %s", "Reward available")
        else
            broker.text = "No keystone"
        end
        return
    end

    if addon.options.showRecordsInBrokerText and character.record then
        local mapId, level = unpack(character.record)
        local instanceName = addon:GetMapName(mapId)
        broker.text = format("%s (Best: %s %s%d|r)", addon:GetKeystoneDisplayText(character.keystone, true), instanceName, addon:GetKeystoneLevelColour(level), level)
        return
    end

    broker.text = addon:GetKeystoneDisplayText(character.keystone, true)
end

addon:RegisterCallback(addon.EVENT_KEYSTONE, UpdateBrokerText)
addon:RegisterCallback(addon.EVENT_OPTION_CHANGED, UpdateBrokerText)
addon:RegisterCallback(addon.EVENT_RECORD, UpdateBrokerText)
